import torch
import torch.nn.functional as F
from typing import List, Union, Iterable
from torch.utils.data import Dataset
from transformers import PreTrainedTokenizer
from saf import Sentence


class TokenizedDataSet(Dataset):
    """
    A wrapper class for acessing a collection of Sentence objects or a list of strings as a torch.utils.data.Dataset


    Args:
        source (Union[Iterable[Sentence], List[str]]): a iterable of SAF Sentence objects (e.g., SentenceDataSet) or list of strings.
        tokenizer (PreTrainedTokenizer): tokenizer object to be used for tokenizing the sentences.
        max_len (int): maximum sentence length.
        one_hot (bool): whether the token ids will be one-hot encoded in the sentence tensors.
    """
    def __init__(self,  source: Union[Iterable[Sentence], List[str]],
                 tokenizer: PreTrainedTokenizer,
                 max_len: int,
                 one_hot: bool = False):
        self.source = source
        self.tokenizer = tokenizer
        self.max_len = max_len
        self.one_hot = one_hot

    def __len__(self):
        return len(self.source)

    def __getitem__(self, idx):
        sentences = self.source[idx]
        if (isinstance(idx, slice)):
            if (isinstance(self.source[0], Sentence)):
                sentences = [sent.surface for sent in sentences]
        else:
            if (isinstance(sentences, Sentence)):
                sentences = [sentences.surface]
            else:
                sentences = [sentences]

        if (not self.tokenizer.pad_token):
            self.tokenizer.pad_token = self.tokenizer.eos_token
            self.tokenizer.pad_token_id = self.tokenizer.eos_token_id

        tokenized = self.tokenizer(sentences, padding="max_length", truncation=True, max_length=self.max_len, return_tensors='pt')

        if (self.one_hot):
            tokenized = F.one_hot(tokenized["input_ids"], num_classes=len(self.tokenizer.get_vocab())).to(torch.int8)
        else:
            tokenized = tokenized["input_ids"]

        return tokenized



from .wiktionary import WiktionaryDefinitionCorpus
from .codwoe import CODWOEDataSet
from .cpae import CPAEDataSet
from .allnli import AllNLIDataSet
from .stsb import STSBDataSet
from .entailmentbank import EntailmentBankDataSet
from .wordnet_filtered import WordNetFilteredDataSet
from .wordnet_spanish import SpanishWordNetFilteredDataSet
from .inference_types import InferenceTypesDataSet



import datasets
from saf_datasets.data_access.dataset import SentenceDataSet


def gen(dataset: SentenceDataSet):
    for sent in dataset:
        entry = {"tokens": list()}
        for token in sent.tokens:
            entry["tokens"].append(token.surface)
            for annot in token.annotations:
                if (annot not in entry):
                    entry[annot] = list()
                entry[annot].append(token.annotations[annot])
        for annot in sent.annotations:
            entry[f"s:{annot}"] = sent.annotations[annot]

        yield entry


def to_hf(dataset: SentenceDataSet) -> datasets.Dataset:
    """Returns a HuggingFace datasets.Dataset wrapper for the given SentenceDataSet"""
    return datasets.Dataset.from_generator(gen, gen_kwargs={"dataset": dataset})




import gzip
import jsonlines
import pickle
from tqdm import tqdm
from saf import Token
from saf import Sentence, Vocabulary
from .dataset import SentenceDataSet, BASE_URL
from .wiktionary import WiktionaryDefinitionCorpus

PATH = "WordNet/WordNet_filtered_data.jsonl.gz"
URL = BASE_URL + "WordNet_filtered_data.jsonl.gz"

ANNOT_RESOURCES = {
    "pos+lemma+ctag+dep+dsr+srl": {
        "path": "WordNet/wordnet_filtered_spacy_dsr_srl.pickle.gz",
        "url": BASE_URL + "wordnet_filtered_spacy_dsr_srl.pickle.gz"
    }
}


class WordNetFilteredDataSet(SentenceDataSet):
    """
    WordNet filtered data set: A collection of filtered definition sentences from WordNet

    Each element of this dataset is a definition sentence (gloss) of a single synset, as it appears in WordNet.
    Definitions were filtered to only include those which definiendum (the term being defined) is representable by
    a single token from the LLaMa tokenizer.
    Each sentence is annotated with a "definiendum" and corresponding metadata from WordNet (WordNet id,
    brown_frequency, wordnet_frequency, category, abstraction_level, generalization_level).

    A pre-annotated instance can be obtained as follows:

    >>> dataset = WordNetFilteredDataSet.from_resource("pos+lemma+ctag+dep+dsr+srl")

    """
    def __init__(self, path: str = PATH, url: str = URL):
        super(WordNetFilteredDataSet, self).__init__(path, url)
        if (not url):
            return

        with gzip.open(self.data_path) as dataset_file:
            with jsonlines.Reader(dataset_file) as reader:
                self.data = list()
                for line in tqdm(reader, desc="Loading WordNet filtered data"):
                    sentence = Sentence()
                    sentence.surface = line["definition"]
                    sentence.annotations["definiendum"] = line["name"]
                    sentence.annotations["brown_frequency"] = line["brown_frequency"]
                    sentence.annotations["wordnet_frequency"] = line["wordnet_frequency"]
                    sentence.annotations["category"] = line["category"]
                    sentence.annotations["abstraction_level"] = line["abstraction_level"]
                    sentence.annotations["generalization_level"] = line["generalization_level"]
                    sentence.annotations['id'] = line["id"]

                    for tok in line["definition_attributes"]:
                        token = Token()
                        token.surface = tok['name']
                        token.annotations = tok #add the ID to the token annotations
                        del token.annotations['name']
                        del token.annotations['definition']
                        sentence.tokens.append(token)

                    self.data.append(sentence)

    def __iter__(self):
        return iter(self.data)

    def __len__(self):
        return len(self.data)

    def __getitem__(self, idx: int) -> Sentence:
        """Fetches the ith definition in the dataset.

        Args:
            idx (int): index for the ith term in the dataset.

        :return: A single term definition (Sentence).
        """
        return self.data[idx]

    def vocabulary(self, source: str = "_token", lowercase: bool = True) -> Vocabulary:
        return WiktionaryDefinitionCorpus.vocabulary(self, source, lowercase)


    @staticmethod
    def from_resource(locator: str):
        """
        Downloads a pre-annotated resource available at the specified locator

        Example:
            >>> dataset = WordNetFilteredDataSet.from_resource("pos+lemma+ctag+dep+dsr+srl")
        """
        dataset = None
        if (locator in ANNOT_RESOURCES):
            path = ANNOT_RESOURCES[locator]["path"]
            url = ANNOT_RESOURCES[locator]["url"]
            data_path = SentenceDataSet.download_resource(path, url)
            with gzip.open(data_path, "rb") as resource_file:
                dataset = pickle.load(resource_file)
        else:
            print(f"No resource found at locator: {locator}")

        return dataset








import gzip
import jsonlines
from tqdm import tqdm
from saf import Token
from saf import Sentence, Vocabulary
from .dataset import SentenceDataSet, BASE_URL


PATH = "WordNet/Spanish_WordNet_filtered_data.jsonl.gz"
URL = BASE_URL + "Spanish_WordNet_filtered_data.jsonl.gz"


class SpanishWordNetFilteredDataSet(SentenceDataSet):
    """
        Spanish WordNet filtered data set: A collection of filtered definition sentences from the Spanish WordNet

        Each element of this dataset is a definition sentence (gloss) of a single synset, as it appears in WordNet.
        Definitions were filtered to only include those which definiendum (the term being defined) is representable by
        a single token from the LLaMa tokenizer.
        Each sentence is annotated with a "definiendum" and corresponding metadata from WordNet (WordNet id,
        cess_esp_frequencies, wordnet_frequency, category, abstraction_level, generalization_level).

    """
    def __init__(self, path: str = PATH, url: str = URL):
        super(SpanishWordNetFilteredDataSet, self).__init__(path, url)
        with gzip.open(self.data_path, 'rt') as dataset_file:
            with jsonlines.Reader(dataset_file) as reader:
                self.data = list()
                for line in tqdm(reader, desc="Loading WordNet filtered data"):
                    if (not line["definition"]):
                        continue

                    sentence = Sentence()
                    sentence.surface = line["definition"]
                    sentence.annotations["definiendum"] = line["name"]
                    sentence.annotations["cess_esp_frequencies"] = line["cess_esp_frequencies"]
                    sentence.annotations["wordnet_frequency"] = line["wordnet_frequency"]
                    sentence.annotations["category"] = line["category"]
                    sentence.annotations["abstraction_level"] = line["abstraction_level"]
                    sentence.annotations["generalization_level"] = line["generalization_level"]
                    sentence.annotations['id'] = line["id"]
                    sentence.annotations['definition_src'] = line['definition_src']

                    for tok in line["definition_attributes"]:
                        token = Token()
                        token.surface = tok['name']
                        token.annotations = tok #add the ID to the token annotations
                        sentence.tokens.append(token)

                    self.data.append(sentence)

    def __iter__(self):
        return iter(self.data)

    def __len__(self):
        return len(self.data)

    def __getitem__(self, idx: int) -> Sentence:
        """Fetches the ith definition in the dataset.

        Args:
            idx (int): index for the ith term in the dataset.

        :return: A single term definition (Sentence).
        """
        return self.data[idx]


import os
import gdown
from typing import Dict, Iterable
from pathlib import Path
from tqdm import tqdm
from zipfile import ZipFile
from torch import argmax, as_tensor, cuda
from transformers import AutoTokenizer, AutoConfig, AutoModelForTokenClassification
from saf import Sentence
from saf.annotators import Annotator
from .models import PRETRAINED_MODELS

BASE_PATH = ".saf_models"

class TransformerAnnotator(Annotator):
    """
    Annotator class for general transformer-based annotation (token classification) models

    Args:
        annot_model (str): the path or identifier for a pretrained model checkpoint.
        tag (str): the annotation tag to which the annotations will be stored (e.g., NER, SRL, DSR).
        labels (Dict[int, str]): valid labels and their corresponding values (ids).
        max_len (int): maximum sentence length.
        device (str): device (e.g., 'cpu', 'cuda') where the model will be stored and ran.
    """
    def __init__(self, annot_model: str, tag: str, labels: Dict[int, str] = None, max_len: int = 128, device: str = None):
        super(TransformerAnnotator, self).__init__()
        if (not device):
            if (cuda.is_available()):
                self.device = "cuda"
            else:
                self.device = "cpu"
        else:
            self.device = device

        if (annot_model in PRETRAINED_MODELS):
            path = PRETRAINED_MODELS[annot_model]["path"]
            url = PRETRAINED_MODELS[annot_model]["url"]
            self.model_path: str = os.path.normpath(os.path.join(str(Path.home()), BASE_PATH, path))
            if (not os.path.exists(self.model_path)):
                os.makedirs(os.path.join(*os.path.split(self.model_path)[:-1]), exist_ok=True)
                gdown.download(url, self.model_path)
                ZipFile(self.model_path).extractall(self.model_path.replace(".zip", ""))
            annot_model = self.model_path.replace(".zip", "")

        self.annot_model = AutoModelForTokenClassification.from_pretrained(annot_model).to(self.device)
        self.annot_tokenizer = AutoTokenizer.from_pretrained(annot_model)
        self.annot_config = AutoConfig.from_pretrained(annot_model)
        self.ids_to_labels = labels if labels else self.annot_config.id2label
        self.max_len = max_len
        self.tag = tag

    def annotate(self, sentences: Iterable[Sentence]):
        for sent in tqdm(sentences, desc="Annotating (Transformer)"):
            #tag each definition
            sentence = sent.surface
            inputs = self.annot_tokenizer(sentence.strip().split(),
                                          is_split_into_words=True,
                                          return_offsets_mapping=True,
                                          truncation=True,
                                          padding='max_length',
                                          max_length=self.max_len)
            item = {key: as_tensor(val) for key, val in inputs.items()}
            ids = item["input_ids"].unsqueeze(0).to(self.device)
            mask = item["attention_mask"].unsqueeze(0).to(self.device)
            # forward pass
            output = self.annot_model(input_ids=ids, attention_mask=mask)
            logits = output.logits
            active_logits = logits.view(-1, self.annot_model.num_labels) # shape (batch_size * seq_len, num_labels)
            flattened_predictions = argmax(active_logits, axis=1) # shape (batch_size*seq_len,) - predictions at the token level
            # mapping tags to words
            tokens = self.annot_tokenizer.convert_ids_to_tokens(ids.squeeze().tolist())
            token_predictions = [self.ids_to_labels[i] for i in flattened_predictions.cpu().numpy()]
            wp_preds = list(zip(tokens, token_predictions)) # list of tuples. Each tuple = (wordpiece, prediction)
            prediction = []
            for token_pred, mapping in zip(wp_preds, item["offset_mapping"].squeeze().tolist()):
                #only predictions on first word pieces are considered
                if mapping[0] == 0 and mapping[1] != 0:
                    prediction.append(token_pred[1])
                else:
                    continue
            word_tags = [[word, label] for word, label in zip(sentence.split(), prediction)]
            #mapping back to the original tokens via string matching
            for i in range(len(sent.tokens)):
                if sent.tokens[i].surface.isalnum():
                    for j in range(len(word_tags)):
                        if sent.tokens[i].surface in word_tags[j][0]:
                            word_tags[j][0] = word_tags[j][0].replace(sent.tokens[i].surface,"")
                            sent.tokens[i].annotations[self.tag] = word_tags[j][1]
                            break
                else:
                    sent.tokens[i].annotations[self.tag] = "O"



import jsonlines
import pickle
import gzip
from typing import Tuple, List, Dict, Union
from tqdm import tqdm
from spacy.lang.en import English
from saf import Sentence, Token, Vocabulary
from saf_datasets.annotators.spacy import SpacyAnnotator
from .dataset import SentenceDataSet, BASE_URL

PATH = "wiktionary/raw-wiktextract-data_2024-01-20.jsonl.gz"  # From https://kaikki.org/dictionary/rawdata.html
URL = "https://drive.google.com/uc?id=1ZMZe25lhjl14p0kEJVxedwmMZV2KUBrz"
ANNOT_RESOURCES = {
    "pos+lemma+ctag+dep+dsr": {
        "path": "wiktionary/wiktdef_spacy_dsr.pickle.gz",
        "url": BASE_URL + "wiktdef_spacy_dsr.pickle.gz"
    },
    "pos+lemma+ctag+dep+dsr#sample": {
        "path": "wiktionary/wiktdef_spacy_dsr_sample.pickle.gz",
        "url": BASE_URL + "wiktdef_spacy_dsr_sample.pickle.gz"
    }
}


class WiktionaryDefinitionCorpus(SentenceDataSet):
    """
    Wiktionary definition corpus: A collection of definition sentences from Wiktionary

    Each element of this dataset is a definition sentence (gloss) of a single sense, as it appears in Wiktionary.
    Each sentence is annotated with a "definiendum" (the term being defined) and its corresponding part-of-speech (definition_pos).

    A pre-annotated instance can be obtained as follows:

    >>> dataset = WiktionaryDefinitionCorpus.from_resource("pos+lemma+ctag+dep+dsr")

    """
    def __init__(self, path: str = PATH, url: str = URL, langs: Tuple[str] = ("English",)):
        super(WiktionaryDefinitionCorpus, self).__init__(path, url)

        self._size: int = -1
        self.langs: Tuple[str] = langs
        self._index: Dict[str, List[Sentence]] = dict()
        self._definitions: List[Sentence] = list()
        self.tokenizer = English().tokenizer

        if (not url):
            return

        self._source = gzip.open(self.data_path)
        self._source.seek(0)

    def __iter__(self):
        if (self._index):
            return iter(self._definitions)
        else:
            self._source.seek(0)
            return WiktionaryDefinitionCorpusIterator(self)

    def __len__(self):
        if (self._size == -1):
            with jsonlines.Reader(self._source) as reader:
                for term in tqdm(reader, desc="Checking data"):
                    if ("word" in term and term["lang"] in self.langs):
                        self._size += 1

            self._source.seek(0)

        return self._size

    def __getitem__(self, item) -> Union[Sentence, List[Sentence]]:
        """Fetches the ith definition in the corpus or all definitions for a given term.

        Args:
            item: (str) for all definitions of the givem term; (int) for the ith definition in the corpus.

        :return: A single definition (Sentence) or list of definitions.
        """
        self.load_index()
        definition = None
        if (isinstance(item, str)):
            if (item in self._index):
                definition = self._index[item]
            else:
                definition = []
        elif (isinstance(item, int) or isinstance(item, slice)):
            definition = self._definitions[item]

        return definition

    def load_index(self):
        """Loads the corpus data and indexes the term definitions."""
        if not self._index:
            self._size = 0
            for definition in tqdm(self, desc="Loading data"):
                term = definition.annotations["definiendum"]
                if (term not in self._index):
                    self._index[term] = list()
                self._index[term].append(definition)
                self._definitions.append(definition)
            self._size = len(self._definitions)

    def vocabulary(self, source: str = "_token", lowercase: bool = True) -> Vocabulary:
        if (source not in self._vocab):
            self._vocab[source] = Vocabulary(self, source=source, lowercase=lowercase)
        if (source == "_token" or source == "lemma"):
            definiendum_vocab = Vocabulary(self, source="definiendum", lowercase=lowercase)
            self._vocab[source].add_symbols(list(definiendum_vocab.symbols))
            self._vocab[source].freqs.update(definiendum_vocab.freqs)
            definiendum_vocab._vocab = self._vocab[source]._vocab
            definiendum_vocab.freqs = self._vocab[source].freqs
            self._vocab["definiendum"] = definiendum_vocab

        return self._vocab[source]

    @staticmethod
    def from_resource(locator: str):
        """
        Downloads a pre-annotated resource available at the specified locator

        Example:
            >>> dataset = WiktionaryDefinitionCorpus.from_resource("pos+lemma+ctag+dep+dsr")
        """
        wiktdef = None
        if (locator in ANNOT_RESOURCES):
            path = ANNOT_RESOURCES[locator]["path"]
            url = ANNOT_RESOURCES[locator]["url"]
            data_path = WiktionaryDefinitionCorpus.download_resource(path, url)
            with gzip.open(data_path, "rb") as resource_file:
                data = pickle.load(resource_file)

            wiktdef = WiktionaryDefinitionCorpus(url="")
            for definition in tqdm(data, desc=f"Loading data from resource: {locator}"):
                term = definition.annotations["definiendum"]
                if (term not in wiktdef._index):
                    wiktdef._index[term] = list()
                wiktdef._index[term].append(definition)
                wiktdef._definitions.append(definition)
            wiktdef._size = len(wiktdef._definitions)

        return wiktdef


class WiktionaryDefinitionCorpusIterator:
    def __init__(self, wiktc: WiktionaryDefinitionCorpus):
        self._wiktc = wiktc
        self._sent_gen = self.sent_generator()

    def __next__(self):
        return next(self._sent_gen)

    def sent_generator(self):
        with jsonlines.Reader(self._wiktc._source) as reader:
            for term in reader:
                if ("word" in term and term["lang"] in self._wiktc.langs):
                    for sense in term["senses"]:
                        if ("glosses" in sense):
                            for gloss in sense["glosses"]:
                                sentence = Sentence()
                                sentence.annotations["definiendum"] = term["word"]
                                sentence.annotations["definition_pos"] = term["pos"]
                                if ("tags" in sense):
                                    sentence.annotations["tags"] = sense["tags"]
                                # if ("categories" in sense):
                                #     sentence.annotations["categories"] = [ct for ct in sense["categories"] if ":" in ct]
                                definition = gloss
                                sentence.surface = definition
                                for tok in self._wiktc.tokenizer(definition):
                                    token = Token()
                                    token.surface = tok.text
                                    sentence.tokens.append(token)

                                if (len(sentence.tokens) == 0):
                                    continue

                                yield sentence

                # for lang in self._wiktc.langs:
                #     for pos in term["langs"][lang]["meanings"]:
                #         for meaning in term["langs"][lang]["meanings"][pos]:
                #             sentence = Sentence()
                #             sentence.annotations["definiendum"] = term["title"].strip()
                #             sentence.annotations["definition_pos"] = pos
                #             definition = meaning["meaning"].replace("</text>", "").strip()
                #             sentence.surface = definition
                #             for tok in self._wiktc.tokenizer(definition):
                #                 token = Token()
                #                 token.surface = tok.text
                #                 sentence.tokens.append(token)
                #
                #             if (len(sentence.tokens) == 0):
                #                 continue
                #
                #             yield sentence


if __name__ == "__main__":
    defs = WiktionaryDefinitionCorpus()

    i = 0
    for sent in defs:
        print("Sent annotations:", sent.annotations)
        print("Token annotations:", [token.surface for token in sent.tokens])
        i += 1
        if (i > 10):
            break

    examples = defs["avocado"]
    annotator = SpacyAnnotator()
    annotator.annotate(examples)
    for example in examples:
        print(f"{example.annotations['definiendum']} ({example.annotations['definition_pos']}): ")
        print(f"{[(token.surface, token.annotations) for token in example.tokens]}")

    print("Corpus size:", len(defs))



import os
import gdown
import torch
from pathlib import Path
from typing import Tuple, List, Dict, Iterable
from collections import Counter
from torch import Tensor
from torch.nn.functional import relu
from tqdm import tqdm
from saf import Sentence, Vocabulary

BASE_PATH = ".saf_data"
BASE_URL = "http://personalpages.manchester.ac.uk/staff/danilo.carvalho/saf_datasets/"


class SentenceDataSet(Iterable[Sentence]):
    """
    Base class for the sentence datasets. Exposes an iterable of SAF Sentence objects.

    Attributes:
        data_path (str): The path where the dataset is stored.
    """
    def __init__(self, path: str, url: str):
        """
        Initializes the SentenceDataSet object by setting up the data path and downloading the data if necessary.

        Args:
            path (str): The subpath to the dataset within the base path.
            url (str): The URL from which to download the dataset if it does not exist locally.
        """
        self.data_path: str = os.path.normpath(os.path.join(str(Path.home()), BASE_PATH, path))
        self._vocab: Dict[str, Vocabulary] = dict()
        self._embeddings: Dict[str, Tensor] = dict()
        self._emb_indices: Dict[str, Tensor] = dict()
        if (not os.path.exists(self.data_path) and url):
            os.makedirs(os.path.join(*os.path.split(self.data_path)[:-1]), exist_ok=True)
            gdown.download(url, self.data_path)

    def __iter__(self):
        raise NotImplementedError

    def __len__(self):
        raise NotImplementedError

    def __getitem__(self, item):
        raise NotImplementedError

    def vocabulary(self, source: str = "_token", lowercase: bool = True) -> Vocabulary:
        """
        Retrieves or creates a Vocabulary object for a specified annotation source.

        Args:
            source (str): The source (annotation) key for which the vocabulary should be retrieved or created (e.g., POS, NER).
            lowercase (bool): Whether to convert all tokens to lowercase in the vocabulary.

        Returns:
            Vocabulary: The vocabulary object for the specified annotation source.
        """
        if (source not in self._vocab):
            self._vocab[source] = Vocabulary(self, source=source, lowercase=lowercase)

        return self._vocab[source]

    def add_symbols(self, symbols: List[str], source: str = "_token"):
        """
        Adds symbols to the vocabulary for a specified annotation source.

        Args:
            symbols (List[str]): A list of symbols to add to the vocabulary.
            source (str): The source (annotation) key for which the symbols should be added.
        """
        vocab = self.vocabulary(source)
        vocab.add_symbols(symbols)

    def del_symbols(self, symbols: List[str], source: str = "_token"):
        """
        Deletes symbols from the vocabulary for a specified annotation source.

        Args:
            symbols (List[str]): A list of symbols to delete from the vocabulary.
            source (str): The source (annotation) key for which the symbols should be deleted.
        """
        vocab = self.vocabulary(source)
        vocab.del_symbols(symbols)

    def to_indices(self, source: str = "_token", default: int = -1, padding: int = 0, pad_symbol: str = None,
                   start_symbol: str = None, end_symbol: str = None) -> Tensor:
        """
        Converts sentences to indices based on the vocabulary for a specified source.

        Args:
            source (str): The source (annotation) key for which the indices should be generated.
            default (int): The default index to use for tokens not found in the vocabulary.
            padding (int): The padding size to apply to the sequences.
            pad_symbol (str): The symbol used for padding.
            start_symbol (str): The symbol used to denote the start of a sequence.
            end_symbol (str): The symbol used to denote the end of a sequence.

        Returns:
            Tensor: A tensor of indices representing the sentences.
        """
        indices = self.vocabulary(source).to_indices(self, default, padding, pad_symbol, start_symbol, end_symbol)
        return torch.tensor(indices, dtype=torch.int32)

    def embeddings(self, tag: str, device: str = "cpu") -> Tensor:
        """
        Retrieves or computes embeddings for a specified type (tag) and stores them.

        Args:
            tag (str): The type for which embeddings should be retrieved or computed (e.g., w2v, electra, bert).
            device (str): The device in which the embeddings should be stored (e.g., 'cpu', 'cuda').

        Returns:
            Tensor: A tensor containing the embeddings for the specified tag.
        """
        if (tag not in self._embeddings):
            if (tag in self[0].annotations):
                embeddings = [sent.annotations[tag] for sent in tqdm(self, desc="Building embedding indices")]
                self._embeddings[tag] = torch.stack(embeddings).to(device)
                for i, sent in enumerate(self):
                    sent.annotations[tag] = self._embeddings[i]
                    sent.annotations[f"{tag}_idx"] = i
                self._emb_indices[tag] = torch.tensor([sent.annotations[f"{tag}_idx"] for sent in self], dtype=torch.int64)
            else:
                embeddings = list()
                for sent in tqdm(self, desc="Building embedding indices"):
                    embeddings.append(torch.stack([tok.annotations[tag] for tok in sent.tokens]))
                self._embeddings[tag] = torch.stack(embeddings).to(device)
                for i, sent in enumerate(self):
                    for j, tok in enumerate(sent.tokens):
                        tok.annotations[tag] = self._embeddings[tag][i][j]
                        tok.annotations[f"{tag}_idx"] = (i, j)

        return self._embeddings[tag]

    def to_positional_indices(self, source: str = "_token", default: int = -1,
                              repetitions: int = 4, omit_default: bool = True) -> Tensor:
        """
        Converts sentences to positional indices based on the vocabulary for a specified annotation source.

        Args:
            source (str): The source (annotation) key for which the positional indices should be generated.
            default (int): The default index to use for tokens not found in the vocabulary.
            repetitions (int): The maximum number of repetitions for each token.
            omit_default (bool): Whether to omit the default index in the positional indices.

        Returns:
            Tensor: A sparse tensor of positional indices.
        """
        indices = self._vocab[source].to_indices(self, default, 0, None, None, None)
        pos_idx = list()
        pos_val = list()
        for i in range(len(indices)):
            rel_pos = torch.tensor(range(1, len(indices[i]) + 1)) / len(indices[i])
            rep_counter = Counter()
            for j in range(len(indices[i])):
                if (indices[i][j] != default or not omit_default):
                    rep_counter.update((indices[i][j],))
                    if (rep_counter[indices[i][j]] <= repetitions):
                        pos_idx.append([i, indices[i][j], rep_counter[indices[i][j]] - 1])
                        pos_val.append(rel_pos[j])

        return torch.sparse_coo_tensor(list(zip(*pos_idx)), pos_val, (len(indices), len(self.vocabulary()), repetitions)).mT.coalesce()

    def from_positional_indices(self, pos_indices: Tensor, source: str = "_token") -> List[List[str]]:
        """
        Converts positional indices back to sentences based on the vocabulary for a specified source.

        Args:
            pos_indices (Tensor): A tensor of positional indices.
            source (str): The source (annotation) key for which the sentences should be reconstructed.

        Returns:
            List[List[str]]: A list of str lists containing the reconstructed sentences.
        """
        vocab = self._vocab[source]
        sentences = list()
        for i in range(pos_indices.shape[0]):
            sorted, indices = torch.sort(pos_indices[i].to_dense().flatten())
            pos_relu = relu(sorted)
            indices = indices[pos_relu.nonzero()].flatten()
            sentences.append([vocab.get_symbol(k % len(vocab)) for k in indices.tolist()])

        return sentences

    def to_positional_encoding(self, source: str = "_token", default: int = -1, repetitions: int = 4, n: int = 1000,
                               omit_default: bool = True, offset: float = 1.0) -> Tuple[Tensor, Tensor]:
        """
        Converts sentences to positional encodings based on the vocabulary for a specified source.

        Args:
            source (str): The source (annotation) key for which the positional encodings should be generated.
            default (int): The default index to use for tokens not found in the vocabulary.
            repetitions (int): The number of positional encoding dimensions per token.
            n (int): The base of the positional encoding formula.
            omit_default (bool): Whether to omit the default index in the positional encodings.
            offset (float): The offset to apply to the positional encodings.

        Returns:
            Tuple[Tensor, Tensor]: A tuple containing the positional encoding tensor and the positional encoding table.
        """
        indices = self._vocab[source].to_indices(self, default, 0, None, None, None)
        pos_idx = list()
        pos_val = list()
        max_len = max([len(idxs) for idxs in indices])
        pe_tbl = torch.zeros(max_len, repetitions)

        for i in range(max_len):
            for j in range(repetitions // 2):
                pe_tbl[i, 2 * j] = torch.sin(i / torch.tensor(pow(n, (2 * j) / repetitions)))
                pe_tbl[i, 2 * j + 1] = torch.cos(i / torch.tensor(pow(n, (2 * j) / repetitions)))

        for i in range(len(indices)):
            rep_counter = Counter()
            for j in range(len(indices[i])):
                if (indices[i][j] != default or not omit_default):
                    rep_counter.update((indices[i][j],))
                    if (rep_counter[indices[i][j]] <= repetitions):
                        pos_idx.append([i, indices[i][j], rep_counter[indices[i][j]] - 1])
                        pos_val.append(pe_tbl[j, rep_counter[indices[i][j]] - 1] + offset)

        penc = torch.sparse_coo_tensor(list(zip(*pos_idx)), pos_val, (len(indices), len(self._vocab[source]), repetitions))

        return penc.mT.coalesce(), pe_tbl

    def from_positional_encoding(self, pos_encoded: Tensor, pe_table: Tensor, offset: float = 1.0,
                                 source: str = "_token") -> List[List[str]]:
        """
        Converts positional encodings back to sentences based on the vocabulary for a specified source.

        Args:
            pos_encoded (Tensor): A tensor of positional encodings.
            pe_table (Tensor): The positional encoding table.
            offset (float): The offset applied to the positional encodings.
            source (str): The source (annotation) key for which the sentences should be reconstructed.

        Returns:
            List[List[str]]: A list of lists containing the reconstructed sentences.
        """
        vocab = self._vocab[source]
        sentences = list()
        for i in range(pos_encoded.shape[0]):
            penc = pos_encoded[i]
            indices = penc._indices().mT.tolist()
            sent = [None] * len(indices)
            for j, k in indices:
                if (penc[j, k] >= offset - 1):
                    dist = torch.abs(pe_table - penc[j, k])
                    pe_k, pe_i = tuple((dist == dist.max()).nonzero()[0].tolist())
                    sent[pe_k] = vocab.get_symbol(k)

            sentences.append([sym for sym in sent if (sym)])

        return sentences

    @staticmethod
    def download_resource(path: str, url: str):
        """
        Downloads a resource from a specified URL to a specified path

        This method is used to download original or preprocessed data that backs the SentenceDataset instances.

        Args:
            path (str): The subpath to the resource within the base path.
            url (str): The URL from which to download the resource.

        Returns:
            str: The normalized path where the resource has been downloaded.
        """
        data_path: str = os.path.normpath(os.path.join(str(Path.home()), BASE_PATH, path))
        if (not os.path.exists(data_path)):
            os.makedirs(os.path.join(*os.path.split(data_path)[:-1]), exist_ok=True)
            gdown.download(url, data_path)

        return data_path









import os
import gzip
from csv import DictReader
from tqdm import tqdm
from spacy.lang.en import English
from saf import Sentence, Token
from saf_datasets.annotators.spacy import SpacyAnnotator
from saf import Sentence, Vocabulary
from .dataset import SentenceDataSet

PATH = "STSB/stsbenchmark.tsv.gz"
URL = "https://sbert.net/datasets/stsbenchmark.tsv.gz"


class STSBDataSet(SentenceDataSet):
    """
    Wrapper for the STSB dataset (Cer et al. 2017): https://aclanthology.org/S17-2001/

    Sentence pairs are split into adjacent entries, sharing the same 'sid'.

    Sentence annotations: id, sid, split, genre, dataset
    """
    def __init__(self, path: str = PATH, url: str = URL):
        super(STSBDataSet, self).__init__(path, url)
        self.tokenizer = English().tokenizer

        with gzip.open(self.data_path, "rt", encoding="utf-8") as dataset_file:
            self.data = list()
            i = -1
            for row in tqdm(dataset_file, desc="Loading STSB data"):
                if (i < 0):
                    i += 1
                    continue

                fields = row.split("\t")
                sentence1 = fields[6]
                sentence2 = fields[7]
                split = fields[0]
                genre = fields[1]
                data_set = fields[2]
                year = fields[3]
                sid = fields[4]
                score = fields[5]
                for sent in [sentence1, sentence2]:
                    sentence = Sentence()
                    sentence.annotations["split"] = split
                    sentence.annotations["genre"] = genre
                    sentence.annotations["dataset"] = data_set
                    sentence.annotations["year"] = year
                    sentence.annotations["sid"] = sid
                    sentence.annotations["score"] = score
                    sentence.annotations["id"] = i
                    sentence.surface = sent.strip()
                    for tok in self.tokenizer(sentence.surface):
                        token = Token()
                        token.surface = tok.text
                        sentence.tokens.append(token)

                    self.data.append(sentence)
                i += 1

    def __iter__(self):
        return iter(self.data)

    def __len__(self):
        return len(self.data)

    def __getitem__(self, idx) -> Sentence:
        """Fetches the ith sentence in the dataset.

        Args:
            idx (int): index for the ith sentence in the dataset.

        :return: A single term decomposition (Sentence).
        """
        return self.data[idx]




from typing import Iterable
from tqdm import tqdm
from allennlp_models import pretrained
from saf import Sentence
from saf.annotators import Annotator


class AllenSRLAnnotator(Annotator):
    """
    Annotator class for Semantic Role Labelling, based on the AllenNLP SRL model (Argument Structure)
    """
    def __init__(self):
        super(AllenSRLAnnotator, self).__init__()
        self.annot_model = pretrained.load_predictor("structured-prediction-srl-bert")

    def annotate(self, sentences: Iterable[Sentence]):
        for sent in tqdm(sentences, desc="Annotating (AllenNLP SRL)"):
            annots = self.annot_model.predict(sent.surface)
            sent.tokens = [tok for tok in sent.tokens if (tok.surface.strip())]
            for i in range(len(sent.tokens)):
                sent.tokens[i].annotations["srl"] = list()
                if (len(annots["verbs"]) > 0):
                    for v_idx in range(len(annots["verbs"])):
                        # print("V_ANNOTS:", annots["verbs"][v_idx])
                        # print([tok.surface for tok in sent.tokens])
                        # print(annots["words"])
                        sent.tokens[i].annotations["srl"].append(annots["verbs"][v_idx]["tags"][i])
                else:
                    sent.tokens[i].annotations["srl"].append("O")



import os
import pickle
import jsonlines
import gzip
from zipfile import ZipFile
from tqdm import tqdm
from spacy.lang.en import English
from saf import Sentence, Token
from .dataset import SentenceDataSet, BASE_URL

FILE_VERSION = "entailment_trees_emnlp2021_data_v3"
PATH = "EntailmentBank/%s.zip" % FILE_VERSION
URL = BASE_URL + "%s.zip" % FILE_VERSION

ANNOT_RESOURCES = {
    "pos+lemma+ctag+dep+srl#noproof": {
        "path": "EntailmentBank/eb_spacy_srl_noproof.pickle.gz",
        "url": BASE_URL + "eb_spacy_srl_noproof.pickle.gz"
    }
}


class EntailmentBankDataSet(SentenceDataSet):
    """
    Wrapper for the EntailmentBank dataset: https://allenai.org/data/entailmentbank

    Context, hypothesis, question, answer and proof sentences for a single entry in the original dataset are split
    adjacently, and can be grouped by their 'id' annotation.

    Sentence annotations: id, task, split, type
    """
    def __init__(self, path: str = PATH, url: str = URL):
        super(EntailmentBankDataSet, self).__init__(path, url)
        self.tokenizer = English().tokenizer
        if (not url):
            return

        with ZipFile(self.data_path) as dataset_file:
            self.data = list()
            for task in ["task_1", "task_2", "task_3"]:
                for split in ["train", "dev", "test"]:
                    split_file = dataset_file.open(os.path.join(FILE_VERSION, "dataset", task, split + ".jsonl"))
                    reader = jsonlines.Reader(split_file)
                    features = dict()
                    for entry in tqdm(reader, desc=f"Loading EntailmentBank: {task} -- {split}"):
                        for key in ["question", "answer", "hypothesis", "proof", "full_text_proof"]:
                            if (key in entry):
                                features[key] = entry[key]
                        for key in entry["meta"]["triples"]:
                            features["context:" + key] = entry["meta"]["triples"][key]
                        for feat in features:
                            sentence = Sentence()
                            sentence.annotations["task"] = task
                            sentence.annotations["split"] = split
                            sentence.annotations["id"] = entry["id"]
                            sentence.annotations["type"] = feat
                            sentence.surface = features[feat] if features[feat] else ""
                            for tok in self.tokenizer(sentence.surface):
                                token = Token()
                                token.surface = tok.text
                                sentence.tokens.append(token)

                            if (sentence.surface):
                                self.data.append(sentence)

    def __iter__(self):
        return iter(self.data)

    def __len__(self):
        return len(self.data)

    def __getitem__(self, idx: int) -> Sentence:
        """Fetches the ith sentence in the dataset.

        Args:
            idx (int): index for the ith sentence in the dataset.

        :return: A single term decomposition (Sentence).
        """
        return self.data[idx]

    @staticmethod
    def from_resource(locator: str):
        """
        Downloads a pre-annotated resource available at the specified locator

        Example:
            >>> dataset = EntailmentBankDataSet.from_resource("pos+lemma+ctag+dep+srl#noproof")
        """
        dataset = None
        if (locator in ANNOT_RESOURCES):
            path = ANNOT_RESOURCES[locator]["path"]
            url = ANNOT_RESOURCES[locator]["url"]
            data_path = SentenceDataSet.download_resource(path, url)
            dataset = EntailmentBankDataSet(url="")
            with gzip.open(data_path, "rb") as resource_file:
                dataset.data = pickle.load(resource_file)
        else:
            print(f"No resource found at locator: {locator}")

        return dataset




import pickle
import gzip
from csv import DictReader
from tqdm import tqdm
from spacy.lang.en import English
from saf import Sentence, Token
from .dataset import SentenceDataSet, BASE_URL

FILE_VERSION = "tr_data_type_amr_op_v0.2"
PATH = "InferenceTypes/%s.csv.gz" % FILE_VERSION
URL = BASE_URL + "%s.csv.gz" % FILE_VERSION

ANNOT_RESOURCES = {
    "pos+lemma+ctag+dep+amr": {
        "path": "InferenceTypes/inftypes_v0.2.pickle.gz",
        "url": BASE_URL + "inftypes_v0.2.pickle.gz"
    }
}


class InferenceTypesDataSet(SentenceDataSet):
    """
    Wrapper for the Inference Types dataset, an annotated subset of
    the EntailmentBank: https://allenai.org/data/entailmentbank

    Premises and conclusion sentences for a single entry in the original dataset are split
    adjacently, and can be grouped by their 'id' annotation.

    Sentence annotations: id, role, type, new_type, type_amr_op
    """
    def __init__(self, path: str = PATH, url: str = URL):
        super(InferenceTypesDataSet, self).__init__(path, url)
        self.tokenizer = English().tokenizer
        if (not url):
            return

        with gzip.open(self.data_path, "rt", encoding="utf-8") as dataset_file:
            self.data = list()
            for row in tqdm(DictReader(dataset_file), desc="Loading inference types data"):
                premise1 = row["premise1"]
                premise2 = row["premise2"]
                conclusion = row["conclusion"]
                for sent, role in [(premise1, "P1"), (premise2, "P2"), (conclusion, "C")]:
                    sentence = Sentence()
                    sentence.annotations["id"] = row["id"]
                    sentence.annotations["role"] = role
                    sentence.annotations["type"] = row["type"]
                    sentence.annotations["new_type"] = row["new_type"]
                    sentence.annotations["type_amr_op"] = row["type_amr_op"]
                    sentence.surface = sent.strip()
                    for tok in self.tokenizer(sentence.surface):
                        token = Token()
                        token.surface = tok.text
                        sentence.tokens.append(token)

                    self.data.append(sentence)

    def __iter__(self):
        return iter(self.data)

    def __len__(self):
        return len(self.data)

    def __getitem__(self, idx: int) -> Sentence:
        """Fetches the ith sentence in the dataset.

        Args:
            idx (int): index for the ith sentence in the dataset.

        :return: A single term decomposition (Sentence).
        """
        return self.data[idx]

    @staticmethod
    def from_resource(locator: str):
        """
        Downloads a pre-annotated resource available at the specified locator

        Example:
            >>> dataset = InferenceTypesDataSet.from_resource("pos+lemma+ctag+dep+amr")
        """
        dataset = None
        if (locator in ANNOT_RESOURCES):
            path = ANNOT_RESOURCES[locator]["path"]
            url = ANNOT_RESOURCES[locator]["url"]
            data_path = SentenceDataSet.download_resource(path, url)
            with gzip.open(data_path, "rb") as resource_file:
                dataset = pickle.load(resource_file)
        else:
            print(f"No resource found at locator: {locator}")

        return dataset




import os
import bz2
import json
import torch
from typing import Tuple, List, Dict, Iterable
from tqdm import tqdm
from tarfile import TarFile
from spacy.lang.en import English
from torch import Tensor
from saf import Sentence, Token
from saf_datasets.annotators.spacy import SpacyAnnotator
from .dataset import SentenceDataSet, BASE_URL

PATH = "CODWOE/CODWOE.tar"
URL = BASE_URL + "CODWOE.tar"


class CODWOEDataSet(SentenceDataSet):
    def __init__(self, path: str = PATH, url: str = URL, langs: Tuple[str] = ("en",)):
        """
        Wrapper for the CODWOE dataset from Semeval-2022 Task 1 (Mickus et al. 2022): https://github.com/TimotheeMickus/codwoe
        """
        super(CODWOEDataSet, self).__init__(path, url)
        TarFile.open(self.data_path).extractall(os.path.join(*self.data_path.split(os.sep)[:-2]))
        path = os.path.join(*self.data_path.split(os.sep)[:-1])

        self._ids: List[str] = list()
        self._glosses: List[str] = list()
        self._emb_char: Tensor = None
        self._emb_electra: Tensor = None
        self._emb_sgns: Tensor = None

        self.split_indices: Dict[str, Tuple[int, int]] = dict()

        emb_char = list()
        emb_electra = list()
        emb_sgns = list()
        for lang in langs:
            prev_split_idx = 0
            for split in ("train", "dev"):
                with bz2.open(os.path.join(path, f"{lang}.{split}.json.bz2"), "rb") as dataset_file:
                    data = json.load(dataset_file)
                    self.split_indices[split] = (prev_split_idx, len(data))
                    prev_split_idx = len(data)
                    for defn in tqdm(data, desc=f"Loading CODWOE data ({lang}.{split})"):
                        self._ids.append(defn["id"])
                        self._glosses.append(defn["gloss"])
                        emb_char.append(defn["char"])
                        emb_electra.append(defn["electra"])
                        emb_sgns.append(defn["sgns"])

        self._emb_char = torch.tensor(emb_char)
        self._emb_electra = torch.tensor(emb_electra)
        self._emb_sgns = torch.tensor(emb_sgns)
        self._size: int = len(self._ids)
        self._index: Dict[str, List[Sentence]] = dict()
        self._definitions: List[Sentence] = list()
        self.tokenizer = English().tokenizer

        del emb_char
        del emb_electra
        del emb_sgns

    def __iter__(self):
        return CODWOEDataSetIterator(self)

    def __len__(self):
        return self._size

    def __getitem__(self, idx: int) -> Sentence:
        """Fetches the ith definition in the dataset or all definitions for a given term.

        :param item: (int) for the ith definition in the dataset.
        :return: A single definition (Sentence).
        """
        return self.def_convert(idx)

    def def_convert(self, idx: int) -> Sentence:
        """Converts CODWOE definition entry to SAF sentence format

        :param idx: (int) for the ith definition in the corpus.
        :return: A single definition (Sentence).
        """
        sentence = Sentence()
        sentence.annotations["id"] = self._ids[idx]
        sentence.annotations["emb_char"] = self._emb_char[idx]
        sentence.annotations["emb_electra"] = self._emb_electra[idx]
        sentence.annotations["emb_sgns"] = self._emb_sgns[idx]
        sentence.surface = self._glosses[idx].strip()
        for tok in self.tokenizer(sentence.surface):
            token = Token()
            token.surface = tok.text
            sentence.tokens.append(token)

        return sentence

    def embeddings(self, tag: str, device: str = "cpu") -> Tensor:
        embeddings = None
        if (tag == "char"):
            self._emb_char = self._emb_char.to(device)
            embeddings = self._emb_char
        elif (tag == "electra"):
            self._emb_electra = self._emb_electra.to(device)
            embeddings = self._emb_electra
        elif (tag == "sgns"):
            self._emb_sgns = self._emb_sgns.to(device)
            embeddings = self._emb_sgns
        else:
            embeddings = super(CODWOEDataSet, self).embeddings(tag, device)

        return embeddings

    def train_indices(self) -> Tuple[int, int]:
        return self.split_indices["train"]

    def dev_indices(self) -> Tuple[int, int]:
        return self.split_indices["dev"]


class CODWOEDataSetIterator:
    def __init__(self, codwoeds: CODWOEDataSet):
        self._codwoeds: CODWOEDataSet = codwoeds
        self._sent_gen = self.sent_generator()

    def __next__(self):
        return next(self._sent_gen)

    def sent_generator(self):
        for i in range(len(self._codwoeds)):
            yield self._codwoeds.def_convert(i)


if __name__ == "__main__":
    defs = CODWOEDataSet()

    print(defs.vocabulary().freqs.most_common())

    exit(0)

    i = 0
    for sent in defs:
        print("Sent annotations:", sent.annotations)
        print("Token annotations:", [token.surface for token in sent.tokens])
        i += 1
        if (i > 10):
            break

    examples = [defs[5], defs[8]]
    annotator = SpacyAnnotator()
    annotator.annotate(examples)
    for example in examples:
        print(f"{example.annotations['id']} (\"{example.surface}\"): ")
        print(f"{[(token.surface, token.annotations) for token in example.tokens]}")

    print("Corpus size:", len(defs))



import os
import bz2
from tqdm import tqdm
from spacy.lang.en import English
from saf import Sentence, Token
from saf_datasets.annotators.spacy import SpacyAnnotator
from saf import Sentence, Vocabulary
from .dataset import SentenceDataSet, BASE_URL
from .wiktionary import WiktionaryDefinitionCorpus

PATH = "CPAE/cpae_definitions.csv.bz2"
URL = BASE_URL + "cpae_definitions.csv.bz2"


class CPAEDataSet(SentenceDataSet):
    """
    Wrapper for the CPAE dataset (Bosc, Vincent. 2018): https://aclanthology.org/D18-1181/
    """
    def __init__(self, path: str = PATH, url: str = URL):
        super(CPAEDataSet, self).__init__(path, url)
        self.tokenizer = English().tokenizer

        with bz2.open(self.data_path, "rt", encoding="utf-8") as dataset_file:
            self.data = list()
            for line in tqdm(dataset_file, desc="Loading CPAE definition data"):
                fields = line.split(";")
                sentence = Sentence()
                sentence.annotations["definiendum"] = fields[2].strip()
                definition = fields[3].strip()
                sentence.surface = definition
                for tok in self.tokenizer(definition):
                    token = Token()
                    token.surface = tok.text
                    sentence.tokens.append(token)

                self.data.append(sentence)

    def __iter__(self):
        return iter(self.data)

    def __len__(self):
        return len(self.data)

    def __getitem__(self, idx: int) -> Sentence:
        """Fetches the ith definition in the dataset.

        Args:
            idx (int): index for the ith term in the dataset.

        :return: A single term definition (Sentence).
        """
        return self.data[idx]

    def vocabulary(self, source: str = "_token", lowercase: bool = True) -> Vocabulary:
        return WiktionaryDefinitionCorpus.vocabulary(self, source, lowercase)



import os
import tarfile
import amrlib
import gdown
from typing import Iterable
from saf import Sentence
from saf.annotators import Annotator

AMRLIB_MODELS_BASE_URL = "https://github.com/bjascob/amrlib-models/releases/download/"
AMRLIB_MODELS = {
    "parse_xfm_bart_large": "parse_xfm_bart_large-v0_1_0/model_parse_xfm_bart_large-v0_1_0.tar.gz",
    "parse_xfm_bart_base": "parse_xfm_bart_base-v0_1_0/model_parse_xfm_bart_base-v0_1_0.tar.gz",
    "parse_spring": "model_parse_spring-v0_1_0/model_parse_spring-v0_1_0.tar.gz",
    "parse_t5": "model_parse_t5-v0_2_0/model_parse_t5-v0_2_0.tar.gz"
}


class AMRAnnotator(Annotator):
    """
        Annotator class for Abstract Meaning Representation (AMR), based on the amrlib library annotation models

        Args:
            annot_model (str): amrlib annotation model to be used. See `https://github.com/bjascob/amrlib-models`
        """
    def __init__(self, annot_model: str = "parse_xfm_bart_large"):
        super(AMRAnnotator, self).__init__()
        amrlib_path = amrlib.__file__.replace("__init__.py", "")
        model_path = os.path.join(amrlib_path, "data", "model_stog")

        if (not os.path.exists(model_path)):
            os.makedirs(os.path.join(amrlib_path, "data"), exist_ok=True)
            model_filename = AMRLIB_MODELS[annot_model].split("/")[-1]
            model_filepath = os.path.join(amrlib_path, "data", model_filename)
            gdown.download(AMRLIB_MODELS_BASE_URL + AMRLIB_MODELS[annot_model], model_filepath, resume=True)
            model_file = tarfile.open(model_filepath, "r:gz")
            model_file.extractall(os.path.join(amrlib_path, "data"))
            os.replace(os.path.join(amrlib_path, "data", model_filename.split(".")[0]),
                       os.path.join(amrlib_path, "data", "model_stog"))

        self.annot_model = amrlib.load_stog_model()

    def annotate(self, sentences: Iterable[Sentence]):
        graphs = self.annot_model.parse_sents([sent.surface for sent in sentences], disable_progress=False)
        for sent, graph in zip(sentences, graphs):
            sent.annotations["amr"] = graph



import os
from distutils.core import setup

PKG_ROOT = os.path.abspath(os.path.dirname(__file__))


def load_requirements() -> list:
    """Load requirements from file, parse them as a Python list"""

    with open(os.path.join(PKG_ROOT, "requirements.txt"), encoding="utf-8") as f:
        all_reqs = f.read().split("\n")
    install_requires = [str(x).strip() for x in all_reqs]

    return install_requires


setup(
    name='saf_datasets',
    version='0.6.10',
    packages=['saf_datasets', 'saf_datasets.annotators', 'saf_datasets.data_access', 'saf_datasets.wrappers'],
    url='',
    author='Danilo S. Carvalho',
    author_email='danilo.carvalho@manchester.ac.uk',
    description='Data set loading and annotation facilities for the Simple Annotation Framework',
    install_requires=load_requirements()
)



import spacy
from typing import Iterable
from tqdm import tqdm
from saf import Sentence
from saf.annotators import Annotator


class SpacyAnnotator(Annotator):
    """
        Annotator class for part-of-speech (POS), lemmatization, constituency tree tagging,
        and dependency grammar tagging, based on the SpaCy library annotation models

        Args:
            annot_model (str): SpaCy annotation model to be used.
        """
    def __init__(self, annot_model: str = "en_core_web_sm"):
        super(SpacyAnnotator, self).__init__()
        self.annot_model = spacy.load(annot_model)

    def annotate(self, sentences: Iterable[Sentence]):
        for sent in tqdm(sentences, desc="Annotating (Spacy)"):
            annots = self.annot_model(sent.surface)
            for i in range(len(sent.tokens)):
                sent.tokens[i].annotations["pos"] = annots[i].pos_
                sent.tokens[i].annotations["lemma"] = annots[i].lemma_
                sent.tokens[i].annotations["dep"] = annots[i].dep_
                sent.tokens[i].annotations["ctag"] = annots[i].tag_



import os
import gzip
from csv import DictReader
from tqdm import tqdm
from spacy.lang.en import English
from saf import Sentence, Token, Vocabulary
from .dataset import SentenceDataSet

PATH = "AllNLI/AllNLI.tsv.gz"
URL = "https://sbert.net/datasets/AllNLI.tsv.gz"


class AllNLIDataSet(SentenceDataSet):
    """
    Wrapper for the AllNLI dataset: https://www.sbert.net/examples/datasets/README.html

    Sentence annotations: id, dataset, split, label
    """
    def __init__(self, path: str = PATH, url: str = URL):
        super(AllNLIDataSet, self).__init__(path, url)
        self.tokenizer = English().tokenizer

        with gzip.open(self.data_path, "rt", encoding="utf-8") as dataset_file:
            self.data = list()
            reader = DictReader(dataset_file, delimiter="\t")
            i = 0
            for row in tqdm(reader, desc="Loading AllNLI data"):
                for sent in [row["sentence1"], row["sentence2"]]:
                    sentence = Sentence()
                    sentence.annotations["split"] = row["split"]
                    sentence.annotations["label"] = row["label"]
                    sentence.annotations["dataset"] = row["dataset"]
                    sentence.annotations["id"] = i
                    sentence.surface = sent
                    for tok in self.tokenizer(sent):
                        token = Token()
                        token.surface = tok.text
                        sentence.tokens.append(token)

                    self.data.append(sentence)
                i += 1

    def __iter__(self):
        return iter(self.data)

    def __len__(self):
        return len(self.data)

    def __getitem__(self, idx: int) -> Sentence:
        """Fetches the ith sentence in the dataset.

        Args:
            idx (int): index for the ith sentence in the dataset.

        :return: A single term decomposition (Sentence).
        """
        return self.data[idx]

